/*

    File: godmode.c

    Copyright (C) 1998-2000  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define FULL_GODMODE
#include <curses.h>
#include <ctype.h>      /* toupper, tolower */
#include <stdlib.h>
#include <strings.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "analyse.h"
#include "lang.h"
#include "godmode.h"
#include "testdisk.h"
#include "fat.h"
#include "interface.h"
//#define ESC 0x1B
#define RO 1
#define RW 0
extern char* nom_os[256];
void init_structure(int nbr_part, t_diskext tab_part[NBR_PART_MAX]);
int ask_structure(int nbr_part, t_diskext tab_part[NBR_PART_MAX]);
void ask_mbr_order(int nbr_part, t_diskext tab_part[NBR_PART_MAX]);
void init_extended_order(int nbr_part, t_diskext tab_part[NBR_PART_MAX]);
int test_structure(int nbr_part, t_diskext tab_part[NBR_PART_MAX]);
int search_part(t_param_disk_cst disk_car,int *nbr_part,t_diskext tab_part[NBR_PART_MAX],int paranoid,int debug,int dump_ind,int fast_mode);
void reduce_structure(int*nbr_part, t_diskext tab_part[NBR_PART_MAX]);
void add_ext_part(t_param_disk_cst disk_car,int*nbr_part, t_diskext tab_part[NBR_PART_MAX]);
int write_mbr(t_param_disk_cst disk_car,int nbr_part, t_diskext tab_part[NBR_PART_MAX],int,int);
int write_all_log(t_param_disk_cst disk_car,int nbr_part, t_diskext tab_part[NBR_PART_MAX],int,int);
void adjust_part_size(t_param_disk_cst,t_diskext);
int can_be_ext(t_diskext);
void only_one_bootable(int nbr_part, t_diskext tab_part[NBR_PART_MAX], int pos);
int diff(char buffer[0x200], char buffer_org[0x200]);
int compar_diskext(const void*pos1, const void*pos2);
int ask_recover_cih(t_param_disk_cst disk_car,int*nbr_part,t_diskext tab_part[NBR_PART_MAX],int debug);

void adjust_part_size(t_param_disk_cst disk_car,t_diskext part)
{
  part->part_size=CHS2SR2(disk_car,&part->end)-CHS2SR2(disk_car,&part->start)+1;
}


void init_structure(int nbr_part, t_diskext tab_part[NBR_PART_MAX])
{
  /* Reconstruit une structure 
   * nbr_prim nombre de partition primaire
   * first_log, last_log : partition etendu du block etendu
   * nbr_log_block : nombre de block de partition etendu
   *                    devrait etre 0 ou 1 */
  int nbr_prim=0, first_log=-1, last_log=-1, nbr_log_block=0,last_prim=-1;
  int i;
  for(i=0;i<nbr_part;i++)
  {
    if(!can_be_ext(tab_part[i]))
    {
      nbr_prim++;
	  last_prim=i;
      if(i>0 && last_log==i-1)
      { nbr_log_block++; first_log=-1;}
    }
    else
	{
	  if(first_log==-1)
		first_log=i;
	  last_log=i;
	}
  }
  /* Verification */
  if(last_log==nbr_part-1)
    nbr_log_block++;
#ifdef DEBUG
  doprintf("\nRes: nbr_prim %d, nbr_log_block %d",
	nbr_prim, nbr_log_block);
#endif
  if(nbr_prim+nbr_log_block>4)
  { /* Probleme */
#ifdef DEBUG
    doprintf("\nProbleme");
#endif
  }
  else if(nbr_log_block>1)
  { /* Choix */
#ifdef DEBUG
    doprintf("\nChoix");
#endif
    for(i=0;i<nbr_part;i++)
    {
      if(!can_be_ext(tab_part[i]))
	tab_part[i]->status=STATUS_PRIM;
    }
  }
  else
  {
    for(i=0;i<nbr_part;i++)
    {
      if(!can_be_ext(tab_part[i]))
	tab_part[i]->status=STATUS_PRIM;
      else
	tab_part[i]->status=STATUS_LOG;
    }
  }
  /* */
  if(nbr_prim==1)
    if(tab_part[last_prim]->disk==0x80)
	tab_part[last_prim]->status=STATUS_PRIM_BOOT;
}

int can_be_ext(t_diskext partition)
{
  return((partition->start.head>0)&&
      (partition->start.cylinder!=0 ||
       partition->start.head!=1 ||
       partition->start.sector!=1));
}

void only_one_bootable(int nbr_part, t_diskext tab_part[NBR_PART_MAX], int pos)
{
  int i;
  if(tab_part[pos]->status==STATUS_PRIM_BOOT)
    for(i=0;i<nbr_part;i++)
    {
      if((i!=pos)&&(tab_part[i]->status==STATUS_PRIM_BOOT))
	tab_part[i]->status=STATUS_PRIM;
    }
}

int ask_structure(int nbr_part, t_diskext tab_part[NBR_PART_MAX])
{
  int i;
  int pos=0;
  int car;
  int quit;
  clear();
  move(0,0);
  doprintf(m_Copyright);
#ifdef DJGPP
  doprintf("Dos version\n\n");
#else
  doprintf("Linux version\n\n");
#endif
  mvaddstr(5,0,m_CHOOSE_PART_TYPE);
  do
  {
    for(i=0;i<nbr_part;i++)
    {
      move(5+2+i,0);
      if(i==pos)
	standout();
      aff_part(0,tab_part[i]);
      if(i==pos)
	standend();
    }
    move(20,0);
    if(test_structure(nbr_part,tab_part))
      doprintf(m_STRUCT_BAD);
    else
      doprintf(m_STRUCT_OK);
    refresh();
    car=getch();
    quit=0;
    switch(car)
    {
      case 0x103:
      case 'A':
	only_one_bootable(nbr_part,tab_part,pos);
	if(--pos<0)
	  pos=nbr_part-1;
	break;
      case 0x102:
      case 'B':
	only_one_bootable(nbr_part,tab_part,pos);
	if(++pos>=nbr_part)
	  pos=0;
	break;
      case 0x105:	 /* right */
      case 'C':
      case '+':
      case ' ':
      case 'l':
	if(tab_part[pos]->status==STATUS_LOG)
	  tab_part[pos]->status=STATUS_DELETED;
	else
	  tab_part[pos]->status++;
	if(tab_part[pos]->status==STATUS_LOG && !can_be_ext(tab_part[pos]))
	  tab_part[pos]->status=STATUS_DELETED;
	break;
      case 0x104:	/* left */
      case 'D':
      case '-':
      case 'S':
	if(tab_part[pos]->status==STATUS_DELETED)
	  tab_part[pos]->status=STATUS_LOG;
	else
	  tab_part[pos]->status--;
	if(tab_part[pos]->status==STATUS_LOG && !can_be_ext(tab_part[pos]))
	  tab_part[pos]->status--;
	break;
      case 'q':
      case '\r':
      case '\n':
      case 'M':
	quit=1;
	break;
      default:
//      doprintf("Touche %02X", car);
	break;
    }
  } while(!quit ||test_structure(nbr_part,tab_part));
  return 0;
}

int test_structure(int nbr_part, t_diskext tab_part[NBR_PART_MAX])
{ /* Return 1 if bad*/
  int nbr_prim=0;
  int nbr_prim_boot=0;
  int nbr_log_block=0;
  int first_log=-1, last_log=-1;
  int err=-1;
  int i;
  for(i=0;i<nbr_part;i++)
  {
    switch(tab_part[i]->status)
    {
      case STATUS_LOG:
	if(nbr_log_block>0)
	  err=i;
	else
	{
	  if(first_log==-1)
	  { first_log=i; last_log=i; }
	  else
	    last_log=i;
	}
	break;
      case STATUS_PRIM_BOOT:
	if(nbr_prim_boot++)
	  return 1;
      case STATUS_PRIM:
	nbr_prim++;
	if(first_log!=-1 && first_log<=last_log)
	  nbr_log_block++;
	break;
      case STATUS_DELETED:
	break;
    }
  }
  if(last_log!=-1 && nbr_log_block==0)
    nbr_log_block=1;
  /* */
  return(err>=0 || nbr_log_block>1 || nbr_log_block+nbr_prim>4);
}

int search_part(t_param_disk_cst disk_car,int *nbr_part,t_diskext tab_part[NBR_PART_MAX], int paranoid, int debug, int dump_ind,int fast_mode)
{
  dword pos_fin;
  int avance=1;
  int ind_stop=FALSE;
  t_diskext partition=(t_diskext)MALLOC(sizeof(*partition));
  partition->disk=disk_car->disk;
  partition->start.cylinder=0;
  partition->start.head=1;
  partition->start.sector=1;
  /* ncurses: getch will be a non-blocking call */
  while((get_SR_part(disk_car,partition)<=get_SR_param_disk(disk_car))&&!ind_stop)
  {
	move(ANALYSE_Y,ANALYSE_X);
    doprintf("Analyse cylinder %5d/%d",partition->start.cylinder,disk_car->CHS.cylinder);
    refresh();
    partition->end.cylinder=partition->start.cylinder;
    partition->end.head=partition->start.head;
    if(search_type(disk_car,partition,debug,dump_ind))
    {
      doprintf(m_READ_ERROR_AT, partition->start.cylinder,partition->start.head,partition->start.sector);
      avance=0;
    }
    else
    switch(partition->part_type)
    {
      case 0:
      case P_EXTENDED:
      case P_EXTENDX:
	avance=0;
	break;
      default:
	{
	  avance=1;
	  pos_fin=get_SR_part(disk_car,partition)-1+partition->part_size-1;
	  dup_t_diskext(tab_part[*nbr_part],partition);
	  tab_part[*nbr_part]->status=STATUS_DELETED;
	  tab_part[*nbr_part]->sect_rel=0;
	  if(debug)
		  aff_part_rapport(1,tab_part[*nbr_part]);
	  if(pos_fin<=get_LBA_param_disk(disk_car))
	  {
	    if(tab_part[*nbr_part]->end.sector!=disk_car->CHS.sector)
	    { /* Propose d'aligner */
	      tab_part[*nbr_part]->end.sector=disk_car->CHS.sector;
	    }
	    if(tab_part[*nbr_part]->end.head!=disk_car->CHS.head)
	    { /* Propose d'aligner ??? */
	      tab_part[*nbr_part]->end.head=disk_car->CHS.head;
	    }
	    adjust_part_size(disk_car,tab_part[*nbr_part]);
	    if(debug)
	      aff_part_rapport(1,tab_part[*nbr_part]);
	    (*nbr_part)++;
	    /* Testing !!!!!!!!!!!!!!!!! */
	    /*
	       if(tab_part[*nbr_part]->start.cylinder!=0 || tab_part[*nbr_part]->part_type==P_NTFS)
	       (*nbr_part)++;
	       else
	       {
	       doprintf("\nSkip");
	       avance=0;
	       }
	     */
	  }
	  else
	  {
	  }
	}
	break;
    }
    /* Avance */
    partition->start.sector=1;
    if(avance && !paranoid)
    {
      partition->start.cylinder=partition->end.cylinder;
      partition->start.head=partition->end.head;
    }
    if(partition->start.head<disk_car->CHS.head && (!fast_mode || partition->start.head<2))
    {
      partition->start.head++;
    }
    else
    { partition->start.cylinder++; partition->start.head=0;}
	if(getch_nodelay()=='S')
	  ind_stop=1;
  }
  printf("\r                                ");
  FREE(partition);
  return 0;
}

void ask_mbr_order(int nbr_part, t_diskext tab_part[NBR_PART_MAX])
{
  int table[nbr_part];
  int table2[nbr_part];
  int nbr_prim=0;
  int i,pos=0;
  int res;
  int car;
  int quit=0;
  /* Initialisation */
  clear();
  move(0,0);
  doprintf(m_Copyright);
#ifdef DJGPP
  doprintf("Dos version\n\n");
#else
  doprintf("Linux version\n\n");
#endif
  mvaddstr(5,0,m_MBR_ORDER);
  for(i=0;i<nbr_part;i++)
  {
    if((tab_part[i]->status==STATUS_PRIM)||(tab_part[i]->status==STATUS_PRIM_BOOT)||(tab_part[i]->status==STATUS_EXT))
    {
      table[nbr_prim++]=i;
      tab_part[i]->order=nbr_prim;
    }
  }
  do
  {
    uint order;
	/* sort table into table2 */
	int part=0;
	res=0;
	for(order=1;order<=4;order++)
	{
	  int nbr=0;
	  for(i=0;i<nbr_prim;i++)
		if(tab_part[table[i]]->order==order)
		{
		  table2[part++]=table[i];
		  nbr++;
		}
	  res|=(nbr>1);
	}
	/* */
    for(i=0;i<nbr_prim;i++)
	{
	  move(5+2+i,0);
	  if(i==pos)
		standout();
	  aff_part(2,tab_part[table2[i]]);
	  if(i==pos)
		standend();
	}
    move(20,0);
    if(res)
      doprintf(m_MBR_ORDER_BAD);
    else
      doprintf(m_MBR_ORDER_GOOD);
    refresh();
    car=getch();
    quit=0;
    switch(car)
	{
	  case 0x103:
	  case 'A':
		if(--pos<0)
		  pos=nbr_prim-1;
		break;
	  case 0x102:
	  case 'B':
		if(++pos>=nbr_prim)
		  pos=0;
		break;
	  case '1':
	  case '2':
	  case '3':
	  case '4':
		tab_part[table2[pos]]->order=car-'0';
		break;
	  case 0x105:	/* Right */
	  case ' ':
	  case 'C':
	  case '+':
	  case 'l':
		if(++tab_part[table2[pos]]->order>4)
		  tab_part[table2[pos]]->order=1;
		break;
	  case 0x104:	/* Left arrow */
	  case 'D':
	  case '-':
	  case 'S':
		if(--tab_part[table2[pos]]->order<1)
		  tab_part[table2[pos]]->order=4;
		break;
	  case 'q':
	  case '\r':
	  case '\n':
	  case 'M':
		quit=1;
		break;
	}
	doprintf("res %d quit %d", res,quit);
	refresh();
  } while(res || !quit);
}

void init_extended_order(int nbr_part, t_diskext tab_part[NBR_PART_MAX])
{ /* Initialisation */
  int nbr_log=0;
  int i;
  for(i=0;i<nbr_part;i++)
  {
    if(tab_part[i]->status==STATUS_LOG)
      tab_part[i]->order=(++nbr_log)+4;
  }
}

void reduce_structure(int *nbr_part, t_diskext tab_part[NBR_PART_MAX])
{ 
  int pos_read,pos_write=0;
  for(pos_read=0;pos_read<*nbr_part;pos_read++)
  {
    if(tab_part[pos_read]->status!=STATUS_DELETED)
    {
	  if(pos_read==pos_write)
		pos_write++;
	  else
	  {
		dup_t_diskext(tab_part[pos_write++], tab_part[pos_read]);
		tab_part[pos_read]->status=STATUS_DELETED;
	  }
    }
  }
  *nbr_part=pos_write;
  /*
  int i=0,j;
  while (i<(*nbr_part))
  {
    if(tab_part[i]->status==STATUS_DELETED)
//      dup_t_diskext(tab_part[i], tab_part[--(*nbr_part)]);
    {
      for(j=i;j<*nbr_part-1;j++)
	dup_t_diskext(tab_part[j], tab_part[j+1]);
      (*nbr_part)--
    }
    else
      i++;
  }
  */
}

void add_ext_part(t_param_disk_cst disk_car,int *nbr_part, t_diskext tab_part[NBR_PART_MAX])
{
  /* tab_part need to be sorted! */
  /* All extended partitions of an P_EXTENDX  are P_EXTENDED */
  int i,deb=-1,fin=-1;
  for(i=0;i<*nbr_part;i++)
  {
	doprintf("add_ext_part %d\n",i);
    if(tab_part[i]->status==STATUS_LOG)
    {
      if(deb==-1) deb=fin=i; else fin=i;
    }
  }
  if(deb!=-1)
  {
    tab_part[*nbr_part]->start.sector=1;
    tab_part[*nbr_part]->disk=tab_part[0]->disk;
    if((*nbr_part)-(fin-deb)==4)
    {
#ifdef DEBUG
      doprintf("4 entrees dans le MBR => Creer une grande partition etendue");
#endif
      tab_part[*nbr_part]->start.cylinder=(deb==0?1:tab_part[deb-1]->end.cylinder+1);
      tab_part[*nbr_part]->start.head=0;
      tab_part[*nbr_part]->end.cylinder=(fin==*nbr_part-1?disk_car->CHS.cylinder:tab_part[fin+1]->start.cylinder-1);
      tab_part[*nbr_part]->end.head=disk_car->CHS.head;
      tab_part[*nbr_part]->end.sector=disk_car->CHS.sector;
    }
    else
    {
#ifdef DEBUG
      doprintf("Creer une partition etendue de taille minimale");
#endif
      tab_part[*nbr_part]->start.cylinder=tab_part[deb]->start.cylinder;
      tab_part[*nbr_part]->start.head=tab_part[deb]->start.head-1;
      dup_t_CHS(&tab_part[*nbr_part]->end,&tab_part[fin]->end);
    }
    tab_part[*nbr_part]->part_type=(tab_part[*nbr_part]->end.cylinder>1023?P_EXTENDX:P_EXTENDED);
    tab_part[*nbr_part]->status=STATUS_EXT;
    adjust_part_size(disk_car,tab_part[*nbr_part]);
    (*nbr_part)++;
  }
#ifdef DEBUG
  refresh();
  getch();
#endif
}

int write_mbr(t_param_disk_cst disk_car,int nbr_part, t_diskext tab_part[NBR_PART_MAX], int ro,int debug)
{
  int i;
  uchar buffer[0x200];
  uchar buffer_org[0x200];
  if(debug)
	ecrit_rapport("\nwrite_mbr: starting...");
  if(read_MBR(disk_car,&buffer_org)) return 1;
  memset((void *)&buffer,0,0x200);
  memcpy(buffer,buffer_org,TAB_PART);
  buffer[0x1FE]=0x55;
  buffer[0x1FF]=0xAA;
  for(i=0;i<nbr_part;i++)
  {
    switch(tab_part[i]->status)
    {
      case STATUS_PRIM:
      case STATUS_PRIM_BOOT:
      case STATUS_EXT:
      case STATUS_EXT_IN_EXT:
	partition2entree(disk_car,1,tab_part[i], 
	    (t_entree*)&buffer[TAB_PART+(tab_part[i]->order-1)*0x10]);
	break;
    }
  }
  if(ro)
  {
    for(i=0;i<4;i++)
      aff_entree(&buffer[0x10*i+TAB_PART]);
#ifdef DUMP
    diff(buffer,buffer_org);
#endif
  }
  else
    return write_MBR(disk_car,&buffer);
  return 0;
}

#ifdef DUMP
int diff(char buffer[0x200], char buffer_org[0x200])
{
  int j;
  if(memcmp(buffer,buffer_org,0x200))
  {
    for(j=0;j<4;j++)
      if(os(&buffer[0x10*j+TAB_PART]))
	aff_entree(&buffer_org[0x10*j+TAB_PART]);
    printf("\nSectors are differents\n");
    for(j=0;j<0x200;j++)
      if(buffer[j]!=buffer_org[j])
	printf("%02X ", j);
  }
  return 0;
}
#endif

int write_all_log(t_param_disk_cst disk_car,int nbr_part, t_diskext tab_part[NBR_PART_MAX], int ro, int debug)
{
  int i,pos_ext=-1;
  int table[nbr_part];
  int nbr_log=0;
  dword pos;
  t_diskext bloc_ext,bloc_nextext;
  if(debug)
	ecrit_rapport("\nwrite_all_log: starting...");
  for(i=0;i<nbr_part;i++)
  {
    switch(tab_part[i]->status)
	{
	  case STATUS_EXT:          /* Pas bootable */
		if((tab_part[i]->part_type==P_EXTENDED) ||
			(tab_part[i]->part_type==P_EXTENDX))
		  pos_ext=i;
		break;
	  case STATUS_LOG:
		table[nbr_log++]=i;
		break;
	}
  }
  if(pos_ext==-1)
  {
    doprintf(m_NO_EXT_PART);
    return 1;
  }
  bloc_ext=(t_diskext)MALLOC(sizeof(*bloc_ext));
  bloc_nextext=(t_diskext)MALLOC(sizeof(*bloc_nextext));
  pos=get_SR_part(disk_car,tab_part[pos_ext]);
  memcpy(bloc_ext,tab_part[pos_ext],sizeof(struct cellule_diskext));
  for(i=0;i<nbr_log;i++)
  {
    /* tabext: partition etendu conteneur */
    /* tab[part[table[i]]: partition */
    int j;
    uchar buffer[0x200];
    uchar buffer_org[0x200];
	if(debug)
	{
	  ecrit_rapport("\nwrite_all_log: D %02X CHS: %d %d %d", bloc_ext->disk, bloc_ext->start.cylinder, bloc_ext->start.head, bloc_ext->start.sector);
	}
    if(read_ext(disk_car,&buffer_org,bloc_ext))
      return 1;
    memset(buffer,0,0x200);
    memcpy(buffer,buffer_org,TAB_PART);
    buffer[0x1FE]=0x55;
    buffer[0x1FF]=0xAA;
    partition2entree(disk_car,get_SR_part(disk_car,bloc_ext),tab_part[table[i]], (t_entree*)&buffer[TAB_PART]);
    if(i<nbr_log-1)
    { /* Construit le pointeur vers la prochaine partition logique */
      bloc_nextext->disk=tab_part[table[i+1]]->disk;
      bloc_nextext->start.cylinder=tab_part[table[i+1]]->start.cylinder;
      bloc_nextext->start.head=tab_part[table[i+1]]->start.head-1;
      bloc_nextext->start.sector=1;
      dup_t_CHS(&bloc_nextext->end,&tab_part[table[i+1]]->end);
      bloc_nextext->part_type=P_EXTENDED;       /* Never P_EXTENDX */
      adjust_part_size(disk_car,bloc_nextext);
      partition2entree(disk_car,pos,bloc_nextext, (t_entree*)&buffer[TAB_PART+0x10]);
    }
    if(ro)
    {
	  for(j=0;j<4;j++)
		if(os(&buffer[0x10*j+TAB_PART]))
		  aff_entree(&buffer[0x10*j+TAB_PART]);
#ifdef DUMP
	  diff(buffer,buffer_org);
#endif
	}
	else
	  write_ext(disk_car,&buffer_org,bloc_ext);
	memcpy(bloc_ext,bloc_nextext,sizeof(struct cellule_diskext));
  }
  FREE(bloc_ext);
  FREE(bloc_nextext);
  return 0;
}

int ask_recover_cih(t_param_disk_cst disk_car,int*nbr_part,t_diskext tab_part[NBR_PART_MAX],int debug)
{ /* tab_part need to be sorted */
  if(*nbr_part==0 || (*nbr_part<NBR_PART_MAX && (tab_part[0]->start.cylinder>0)))
  {
	tab_part[*nbr_part]->disk=disk_car->disk;
	tab_part[*nbr_part]->part_type=0;
	tab_part[*nbr_part]->start.cylinder=0;
	tab_part[*nbr_part]->start.head=1;
	tab_part[*nbr_part]->start.sector=1;
	if(*nbr_part==0)
	  tab_part[*nbr_part]->end.cylinder=disk_car->CHS.cylinder;
	else
	  tab_part[*nbr_part]->end.cylinder=tab_part[0]->start.cylinder-1;
	tab_part[*nbr_part]->end.head=disk_car->CHS.head;
	tab_part[*nbr_part]->end.sector=disk_car->CHS.sector;
    adjust_part_size(disk_car,tab_part[*nbr_part]);
	{
	  if(recover_FAT(disk_car,tab_part[*nbr_part],debug)==0)
	  {
		(*nbr_part)++;
	  }
	}
  }
  return 0;
}

int compar_diskext(const void*pos1, const void*pos2)
{
  t_diskext_cst a=pos1;
  t_diskext_cst b=pos2;
  return compare_CHS(&a->start,&b->start);
}

int god_mode(t_param_disk_cst disk_car,int paranoid,int debug,int dump_ind, int fast_mode)
{
  int   nbr_part=0;
  int   i;
  t_diskext tab_part[NBR_PART_MAX];
  for(i=0;i<NBR_PART_MAX;i++)
    tab_part[i]=(t_diskext)MALLOC(sizeof(struct cellule_diskext));
  if(search_part(disk_car,&nbr_part,tab_part,paranoid,debug,dump_ind,fast_mode))
    return 1;   /* Err de lecture */
//  ask_recover_cih(&nbr_part, tab_part,debug);
  init_structure(nbr_part, tab_part);
  /* Ajoute dans le fichier de log les partitions trouvees */
  ecrit_rapport("\nResults");
  for(i=0;i<nbr_part;i++)
    aff_part_rapport(1,tab_part[i]);
  doprintf(m_GOD_MODE);
  ask_structure(nbr_part,tab_part);
  reduce_structure(&nbr_part,tab_part);
  /* Sort tab_part */
  qsort(&tab_part,nbr_part,sizeof(t_diskext),compar_diskext);
  /* Creer la partition etendue globale */
  add_ext_part(disk_car,&nbr_part,tab_part);
  /* Demande l'ordre des entrees dans le MBR */
  ask_mbr_order(nbr_part,tab_part);
  /* Demande l'ordre des partitions etendues */
  init_extended_order(nbr_part,tab_part);
  if(interface_write(nbr_part,tab_part))
  {
    write_mbr(disk_car,nbr_part,tab_part,RW,debug);
    write_all_log(disk_car,nbr_part,tab_part,RW,debug);
  }
  /* Libere la memoire */
  if(debug)
    ecrit_rapport("god_mode: free the memory\n");
  for(i=0;i<NBR_PART_MAX;i++)
    FREE(tab_part[i]);
  return 0;
}
