#include <stdio.h>
#include <stdlib.h>
#include <curses.h>
#include <string.h>
#include <ctype.h>
#include "types.h"
#include "common.h"
#include "lang.h"
#include "interface.h"
#include "godmode.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "lang.h"
extern char* nom_os[256];
int get_string(char *str, int len, char *def);
void interface_analyse(t_param_disk_cst param_disk, const int debug);
void interface_recovery(t_param_disk_cst param_disk,const int paranoid, const int debug, const int dump_ind, const int fast_mode);
void interface_options(int *paranoid,int *dump_ind,int *fast_mode);
#define GS_DEFAULT -1
#define GS_ESCAPE -2


int get_string(char *str, int len, char *def)
{
    char c;
    int i = 0;
    int x, y;
    int use_def = FALSE;

    getyx(stdscr, y, x);
    clrtoeol();

    str[0] = 0;

    if (def != NULL) {
	mvaddstr(y, x, def);
	move(y, x);
	use_def = TRUE;
    }

    refresh();
    while ((c = getch()) != '\n' && c != CR) {
	switch (c) {
	case ESC:
	    move(y, x);
	    clrtoeol();
	    refresh();
	    return GS_ESCAPE;
	case DEL:
	case '\b':
	    if (i > 0) {
		str[--i] = 0;
		mvaddch(y, x+i, ' ');
		move(y, x+i);
	    } else if (use_def) {
		clrtoeol();
		use_def = FALSE;
	    } else
		putchar(BELL);
	    break;
	default:
	    if (i < len && isprint(c)) {
		mvaddch(y, x+i, c);
		if (use_def) {
		    clrtoeol();
		    use_def = FALSE;
		}
		str[i++] = c;
		str[i] = 0;
	    } else
		putchar(BELL);
	}
	refresh();
    }

    if (use_def)
	return GS_DEFAULT;
    else
	return i;
}

int getch_nodelay()
{
  int res;
  nodelay(stdscr,1);
  res=toupper(getch());
  nodelay(stdscr,0);
  return res;
}

/*
 * Actual function which prints the button bar and highlights the active button
 * Should not be called directly. Call function menuSelect instead.
 */

int menuUpdate( int y, int x, const struct MenuItem *menuItems, const int itemLength, const char *available, const int menuType, int current )
{
    int i, lmargin = x, ymargin = y;
    const char *mcd;

    /* Print available buttons */
    move( y, x ); clrtoeol();
    for( i = 0; menuItems[i].key; i++ )
    {
	char buff[80];
	int lenName;
		const char *mi;

	/* Search next available button */
	while( menuItems[i].key && !strchr(available, menuItems[i].key) )
	{
	    i++;
	}
	if( !menuItems[i].key ) break; /* No more menu items */

	/* If selected item is not available and we have bypassed it,
	   make current item selected */
	if( current < i && menuItems[current].key < 0 ) current = i;

	/* If current item is selected, highlight it */
	if( current == i ) /*attron( A_REVERSE )*/ standout ();

	/* Print item */
	mi = menuItems[i].name;
	lenName = strlen( mi );
#ifdef DJGPP
	if(lenName > itemLength)
	{
	  if( menuType & MENU_BUTTON )
		sprintf(buff, "[%s]",mi);
	  else
		sprintf(buff, "%s",mi);
	}
	else
	{
	  if( menuType & MENU_BUTTON )
		sprintf( buff, "[%*s%-*s]", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	  else
		sprintf( buff, "%*s%-*s", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	}
#else
	if(lenName > itemLength)
	{
	  if( menuType & MENU_BUTTON )
		snprintf(buff, sizeof(buff),"[%s]",mi);
	  else
		snprintf(buff, sizeof(buff),"%s",mi);
	}
	else
	{
	  if( menuType & MENU_BUTTON )
		snprintf( buff, sizeof(buff),"[%*s%-*s]", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	  else
		snprintf( buff, sizeof(buff),"%*s%-*s", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	}
#endif
	mvaddstr( y, x, buff );

	/* Lowlight after selected item */
	if( current == i ) /*attroff( A_REVERSE )*/ standend ();

	/* Calculate position for the next item */
	if( menuType & MENU_VERT )
	{
	    y += 1;
	    if( y >= WARNING_START )
	    {
		y = ymargin;
		x += itemLength + MENU_SPACING;
		if( menuType & MENU_BUTTON ) x += 2;
	    }
	}
	else
	{
	    x += itemLength + MENU_SPACING;
	    if( menuType & MENU_BUTTON ) x += 2;
	    if( x > COLUMNS - lmargin - 12 )
	    {
		x = lmargin;
		y ++ ;
	    }
	}
    }
    /* Print the description of selected item */
    mcd = menuItems[current].desc;
    mvaddstr( WARNING_START + 1, (COLUMNS - strlen( mcd )) / 2, mcd );
    return y;
}

/* This function takes a list of menu items, lets the user choose one *
 * and returns the value keyboard shortcut of the selected menu item  */

int menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
		const char *available, int menuType, int menuDefault )
{
    int i, ylast = y, key = 0, current = menuDefault;
    if( !( menuType & ( MENU_HORIZ | MENU_VERT ) ) )    
    {
	doprintf("Menu without direction. Defaulting horizontal.");
	menuType |= MENU_HORIZ;
    }
    /* Make sure that the current is one of the available items */
    while( !strchr(available, menuItems[current].key) )
    {
	current ++ ;
	if( !menuItems[current].key ) current = 0;
    }
    /* Repeat until allowable choice has been made */
    while( !key )
    {
	/* Display the menu */
	ylast = menuUpdate( y, x, menuItems, itemLength, available,
			    menuType, current );
	refresh();
	key = getch();

	/* Clear out all prompts and such */
//        clear_warning();
	for( i = y; i < ylast; i ++ )
	{
	    move( i, x );
	    clrtoeol();
	}
	move( WARNING_START + 1, 0 );
	clrtoeol();
	/* Cursor keys */
#ifdef DJGPP
		switch(key)
#else
	if( key == ESC )
	{
	    /* Check whether this is a real ESC or one of extended keys */
	    /*nodelay(stdscr, TRUE);*/
	    key = getch();
	    /*nodelay(stdscr, FALSE);*/
	    if( key == /*ERR*/ ESC )
	    {
		/* This is a real ESC */
		key = ESC;
	    }
	    if( key == '[' )
	    {
		/* This is one extended keys */
		switch( getch() )
#endif
		{
		  case 'A': /* Up arrow */
		  case 0x103:
			if( menuType & MENU_VERT )
			{
			  do {
				current -- ;
				if( current < 0 ) while( menuItems[current+1].key ) current ++ ;
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_UP;
			break;
		  case 'B': /* Down arrow */
		  case 0x102:
			if( menuType & MENU_VERT )
			{
			  do {
				current ++ ;
				if( !menuItems[current].key ) current = 0 ;
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_DOWN;
			break;
		  case 'C': /* Right arrow */
		  case 0x105:
			if( menuType & MENU_HORIZ )
			{
			  do {
				current ++ ;
				if( !menuItems[current].key ) 
				{
				  current = 0 ;
				}
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_RIGHT;
			break;
		  case 'D': /* Left arrow */
		  case 0x104:
			if( menuType & MENU_HORIZ )
			{
			  do {
				current -- ;
				if( current < 0 )
				{
				  while( menuItems[current + 1].key ) current ++ ;
				}
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_LEFT;
			break;
		}
#ifndef DJGPP
	    }
	}
#endif
	/* Enter equals to the keyboard shortcut of current menu item */
	if((key==13) || (key==10))
	{
	    key = menuItems[current].key;
	}
	/* Should all keys to be accepted? */
	if( key && (menuType & MENU_ACCEPT_OTHERS) ) break;
	/* Is pressed key among acceptable ones */
	if( key && (strchr(available, tolower(key)) || strchr(available, key)))
	    break;
	/* The key has not been accepted so far -> let's reject it */
	if( key )
	{
		move(5,0);
		doprintf("key %03X",key);
	    key = 0;
	    putchar( BELL );
//            print_warning("Illegal key");
	}
    }
    /* Clear out prompts and such */
//    clear_warning();
    for( i = y; i <= ylast; i ++ )
    {
	move( i, x );
	clrtoeol();
    }
    move( WARNING_START + 1, 0 );
    clrtoeol();
    return key;
}

/* A function which displays "Press a key to continue"                  *
 * and waits for a keypress.                                            *
 * Perhaps calling function menuSelect is a bit overkill but who cares? */

void menuContinue(void)
{
    static struct MenuItem menuContinueBtn[]=
    {
	{ 'c', "", "Press a key to continue" },
	{ 0, NULL, NULL }
    };

    menuSelect(23, 0,
	menuContinueBtn, 0, "c", MENU_HORIZ | MENU_ACCEPT_OTHERS, 0 );
}

/* Function menuSelect takes way too many parameters  *
 * Luckily, most of time we can do with this function */

int menuSimple(struct MenuItem *menuItems, int menuDefault)
{
    int i, j, itemLength = 0;
    char available[MENU_MAX_ITEMS];

    for(i = 0; menuItems[i].key; i++)
    {
	j = strlen(menuItems[i].name);
	if( j > itemLength ) itemLength = j;
	available[i] = menuItems[i].key;
    }
    available[i] = 0;
    return menuSelect(18, 0, menuItems, itemLength, available, MENU_HORIZ | MENU_BUTTON, menuDefault);
}

/* End of command menu support code */

void change_geometry(t_param_disk param_disk)
{
  int done = FALSE;
  char def[LINE_LENGTH];
  char response[LINE_LENGTH];
  int tmp_val;
  int command;

  while (!done) {
	static struct MenuItem menuGeometry[]=
	{
	  { 'c', "Cylinders", "Change cylinder geometry" },
	  { 'h', "Heads", "Change head geometry" },
	  { 's', "Sectors", "Change sector geometry" },
	  { 'd', "Done", "Done with changing geometry" },
	  { 0, NULL, NULL }
	};
	move(COMMAND_LINE_Y, COMMAND_LINE_X);
	clrtoeol();
	refresh();
	command=menuSimple(menuGeometry, 3);
	command=toupper(command);
	switch (command) {
	  case 'C':
		sprintf(def, "%d", param_disk->CHS.cylinder+1);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of cylinders: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_CYLINDERS) {
			param_disk->CHS.cylinder = tmp_val-1;
		  } else
			doprintf("Illegal cylinders value");
		}
		break;
	  case 'H':
		sprintf(def, "%d", param_disk->CHS.head+1);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of heads: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_HEADS) {
			param_disk->CHS.head = tmp_val-1;
		  } else
			doprintf("Illegal heads value");
		}
		break;
	  case 'S':
		sprintf(def, "%d", param_disk->CHS.sector);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of sectors per track: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_SECTORS) {
			param_disk->CHS.sector = tmp_val;
		  } else
			doprintf("Illegal sectors value");
		}
		break;
	  case ESC:
	  case 'D':
		done = TRUE;
		break;
	  default:
		putchar(BELL);
		break;
	}
  }
}

#ifdef OLD
void change_geometry(t_param_disk param_disk)
{
  unsigned int val;
  doprintf("\nCylinder ?");
  refresh();
  fscanf(stdin,"%u",&val);
  param_disk->CHS.cylinder=val-1;
  doprintf("Head ?");
  refresh();
  fscanf(stdin,"%u",&val);
  param_disk->CHS.head=val-1;
  doprintf("Sector ?");
  refresh();
  fscanf(stdin,"%u",&val);
  param_disk->CHS.sector=val;
}
#endif

void do_curses_testdisk(int debug,int paranoid,int dump_ind, int fast_mode)
{
  int i;
  int command;
  int nbr_disk;
  int current_disk=0;
  int done=0;
  t_param_disk param_disk[NBR_DISK_MAX];
//  t_diskext tab_part[NBR_PART_MAX];
  static struct MenuItem menuMain[]=
  {
	{'a',"Analyse","Analyse current partition structure and recover lost partitions"},
	{'d',"Delete","Delete all partitions"},
	{'c',"MBR Code","Write a classic MBR code"},
	{'g',"Geometry", "Change disk geometry (expert only)" },
	{'o',"Options","Modify options"},
	{'q',"Quit","Quit program"},
	{0,NULL,NULL}
  };
  initscr();
  noecho();
#ifdef DJGPP
  keypad(stdscr, TRUE); // Need it to get arrow key
#else
  nonl(); //don't use for Dos version but enter will work with it... dilema
#endif
  crmode();
  //  intrflush(stdscr, FALSE);
  cbreak();
  for(i=0;i<NBR_DISK_MAX;i++)
  {
	param_disk[i]=(t_param_disk)MALLOC(sizeof(*param_disk[i]));
#ifndef DJGPP
	param_disk[i]->name=NULL;
#endif
  }
  nbr_disk=hd_parse(param_disk,debug);
  while(!done)
  {
	clear();
	move(0,0);
	doprintf(m_Copyright);
#ifdef DJGPP
	doprintf("Dos version\n\n");
#else
	doprintf("Linux version\n\n");
#endif
	move(5,0);
	for(i=0;i<nbr_disk;i++)
	{
	  if(i!=current_disk)
		aff_disk(param_disk[i]);
	  else
	  {
		standout();
		aff_disk(param_disk[i]);
		standend();
	  }
	}
	refresh();
	command = menuSelect(INTER_MAIN_Y, INTER_MAIN_X, menuMain, 8,
		"agdcoq", MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 0);
	switch(command)
	{
	  case 'g':
	  case 'G':
		change_geometry(param_disk[current_disk]);
		break;
	  case 'd':
	  case 'D':
		move(INTER_MAIN_Y,INTER_MAIN_X);
		write_clean_table(param_disk[current_disk]);
		break;
	  case 'c':
	  case 'C':
		move(INTER_MAIN_Y,INTER_MAIN_X);
		write_MBR_code(param_disk[current_disk]);
		break;
	  case 'a':
	  case 'A':
		interface_analyse(param_disk[current_disk],debug);
		interface_recovery(param_disk[current_disk],paranoid,debug, dump_ind, fast_mode);
		break;
	  case 'q':
	  case 'Q':
		done = 1;
		break;
	  case 'o':
	  case 'O':
		interface_options(&paranoid,&dump_ind,&fast_mode);
		break;
	  case MENU_UP:
		if(current_disk>0)
		  current_disk--;
		break;
	  case MENU_DOWN:
		if(current_disk<nbr_disk-1)
		  current_disk++;
		break;
#ifdef DEBUG
	  default:
		move(10,0);
		doprintf("key %03X %d", command, command);
		refresh();
		break;
#endif
	}
  }
  endwin();
  if(debug)
	ecrit_rapport("main: free the memory\n");
  for(i=0;i<NBR_DISK_MAX;i++)
  {
#ifndef DJGPP
	if(param_disk[i]->name)
	  FREE(param_disk[i]->name);
#endif
	FREE(param_disk[i]);    /* forget name */
  }
}

void interface_analyse(t_param_disk_cst param_disk, const int debug)
{
  clear();
  move(0,0);
  doprintf(m_Copyright);
#ifdef DJGPP
  doprintf("Dos version\n\n");
#else
  doprintf("Linux version\n\n");
#endif
  move(5,0);
  doprintf("Check current partition structure");
  move(6,0);
  refresh();
  test_MBR(param_disk,debug);
}

void interface_recovery(t_param_disk_cst param_disk,const int paranoid, const int debug, const int dump_ind, const int fast_mode)
{
  clear();
  move(0,0);
  doprintf(m_Copyright);
#ifdef DJGPP
  doprintf("Dos version\n\n");
#else
  doprintf("Linux version\n\n");
#endif
  move(5,0);
  god_mode(param_disk,paranoid,debug,dump_ind,fast_mode);
}

int interface_write(const int nbr_part,t_diskext_cst tab_part[NBR_PART_MAX])
{
  int i;
  clear();
  move(0,0);
  doprintf(m_Copyright);
#ifdef DJGPP
  doprintf("Dos version\n\n");
#else
  doprintf("Linux version\n\n");
#endif
  move(5,0);
  aff_buffer(BUFFER_RESET,"");
  for(i=0;i<nbr_part;i++)
	if(tab_part[i]->status!=STATUS_LOG)
	  aff_part_buffer(1,tab_part[i]);
  for(i=0;i<nbr_part;i++)
	if(tab_part[i]->status==STATUS_LOG)
	  aff_part_buffer(1,tab_part[i]);
  return aff_buffer(BUFFER_DISPLAY,"W");
}

void interface_options(int *paranoid,int *dump_ind,int *fast_mode)
{
/* paranoid: search in partition
 * dump_ind dump sector Y/N
 * fast_mode Partitions are aligned on cylinder boundary Y/N
 * */
  int done = FALSE;
  int menu = 3;
  // write old options to log file
  ecrit_rapport("Old options :\n%s\n%s\n%s\n",
	*paranoid?"Paranoid : Yes":"Paranoid : No",
	*dump_ind?"Dump : Yes":"Dump : No",
	*fast_mode?"Cylinder boundary : Yes":"Cylinder boundary : No");
  // ncurses interface
  while (!done) {
	struct MenuItem menuOptions[]=
	{
	  { 'P', NULL, "Search partition" },
	  { 'D',NULL,"Dump essential sectors" },
	  { 'C',NULL,"Partitions are aligned on cylinder boundary" },
	  { 'Q',"Quit","Return to main menu"},
	  { 0, NULL, NULL }
	};
	menuOptions[0].name=*paranoid?"Paranoid : Yes":"Paranoid : No";
	menuOptions[1].name=*dump_ind?"Dump : Yes":"Dump : No";
	menuOptions[2].name=*fast_mode?"Cylinder boundary : Yes":"Cylinder boundary : No";
//      move(COMMAND_LINE_Y, COMMAND_LINE_X);
//      clrtoeol();
	refresh();
    switch (toupper( menuSelect(INTER_OPTION_Y, INTER_OPTION_X, menuOptions, 0, "PDCQ", MENU_VERT, menu)))
	{
	  case 'P':
		*paranoid=!*paranoid;
		menu=0;
		break;
	  case 'D':
		*dump_ind=!*dump_ind;
		menu=1;
		break;
	  case 'C':
		*fast_mode=!*fast_mode;
		menu=2;
		break;
	  case ESC:
	  case 'Q':
		done = TRUE;
		break;
	}
  }
  // write new options to log file
  ecrit_rapport("New options :\n%s\n%s\n%s\n",
	*paranoid?"Paranoid : Yes":"Paranoid : No",
	*dump_ind?"Dump : Yes":"Dump : No",
	*fast_mode?"Cylinder boundary : Yes":"Cylinder boundary : No");
}

void dump(const uchar *nom_dump,uint lng)
{
  int i,j;
  uchar car;
  int done=0;
  int menu=2;   /* default : quit */
  int pos=0;
  struct MenuItem menuDump[]=
  {
	{ 'P', "Previous",""},
	{ 'N', "Next","" },
	{ 'Q',"Quit","Quit dump section"},
	{ 0, NULL, NULL }
  };
  /* write dump to log file*/
  for (i=pos; (i<lng/0x10)&&((i-pos)<DUMP_MAX_LINES); i++)
  {
	ecrit_rapport("%04X ",i*0x10);
	for(j=0; j< 0x10;j++)
	{
	  car=*(nom_dump+i*0x10+j);
	  ecrit_rapport("%02x", car);
	  if(j%4==(4-1))
		ecrit_rapport(" ");
	}
	ecrit_rapport("  ");
	for(j=0; j< 0x10;j++)
	{
	  car=*(nom_dump+i*0x10+j);
	  if ((car<30)||(car > 127))
		ecrit_rapport(" ");
	  else
		ecrit_rapport("%c",  car);
	}
	ecrit_rapport("\n");
  }
  /* ncurses interface */
  move(DUMP_Y,DUMP_X);
  doprintf(m_DUMP_HEXA);
  do
  {
	for (i=pos; (i<lng/0x10)&&((i-pos)<DUMP_MAX_LINES); i++)
	{
	  move(DUMP_Y+1+i-pos,DUMP_X);
	  doprintf("%04X ",i*0x10);
	  for(j=0; j< 0x10;j++)
	  {
		car=*(nom_dump+i*0x10+j);
		doprintf("%02x", car);
		if(j%4==(4-1))
		  doprintf(" ");
	  }
	  doprintf("  ");
	  for(j=0; j< 0x10;j++)
	  {
		car=*(nom_dump+i*0x10+j);
		if ((car<30)||(car > 127))
		  doprintf(" ");
		else
		  doprintf("%c",  car);
	  }
	}
	switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuDump, 8, "PNQ", MENU_HORIZ | MENU_ACCEPT_OTHERS, menu)))
	{
	  case 'P':
	  case MENU_UP:
		menu=0;
		if(pos>0)
		  pos--;
		break;
	  case 'N':
	  case MENU_DOWN:
		menu=1;
		if(pos<lng/0x10-DUMP_MAX_LINES+1)
		  pos++;
		break;
	  case ESC:
	  case 'Q':
		done = TRUE;
		break;
	}
  } while(!done);
}

int aff_buffer(const int cmd, const char *_format, ...)
{
  static char buffer[MAX_LINES][LINE_LENGTH];
  static int nbr_lines=0;
  switch(cmd)
  {
	case BUFFER_RESET:
	  {
		int i;
		nbr_lines=0;
		for(i=0;i<MAX_LINES;i++)
		  buffer[i][0]='\0';
	  }
	  break;
	case BUFFER_ADD:
	  {
		char tmp_line[LINE_LENGTH];
		char *pos_in_tmp_line=tmp_line-1;
		tmp_line[0]='\0';
#ifdef DJGPP
		vsprintf(tmp_line,_format,(&_format)+1);
#else
		vsnprintf(tmp_line,LINE_LENGTH,_format,(&_format)+1);
#endif
		refresh();
		while(pos_in_tmp_line)
		{
		  char *ret_ligne= strchr(++pos_in_tmp_line,'\n');
		  if(ret_ligne)
			strncat(buffer[nbr_lines++],pos_in_tmp_line,
				ret_ligne-pos_in_tmp_line<LINE_LENGTH?ret_ligne-pos_in_tmp_line:LINE_LENGTH);
		  else
			strncat(buffer[nbr_lines],pos_in_tmp_line,LINE_LENGTH);
		  pos_in_tmp_line=ret_ligne;
		}
	  }
	  break;
	case BUFFER_DISPLAY:
	  {
		int i;
		int menu;
		int pos=0;
		int done=0;
		struct MenuItem menuBuffer[]=
		{
		  { 'P', "Previous",""},
		  { 'N', "Next","" },
		  { 'Q',"Quit","Quit this section"},
		  { 'W', "Write","Write partition structure to disk"},
		  { 0, NULL, NULL }
		};
		char options[20];
		if(nbr_lines<=DUMP_MAX_LINES)
		{
		  strcpy(options,"Q");
		  menu=0;       /* default = quit */
		}
		else
		{
		  strcpy(options,"PNQ");
		  menu=1;       /* default = next */
		}
		strcat(options,_format);
		if(buffer[nbr_lines][0])
		  nbr_lines++;
		do
		{
		  for (i=pos; (i<nbr_lines)&&((i-pos)<DUMP_MAX_LINES); i++)
		  {
			move(DUMP_Y+1+i-pos,DUMP_X);
			doprintf("%s",buffer[i]);
		  }
		  
		  switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuBuffer, 8, options, MENU_HORIZ | MENU_ACCEPT_OTHERS, menu)))
		  {
			case 'P':
			case MENU_UP:
			  menu=0;
			  if(pos>0)
				pos--;
			  break;
			case 'N':
			case MENU_DOWN:
			  menu=1;
			  if(pos<nbr_lines-DUMP_MAX_LINES+1)
				pos++;
			  break;
			case 'W':
			  if(strchr(options,'W'))
				return 1;
			case ESC:
			case 'Q':
			  done = TRUE;
			  break;
		  }
		} while(!done);
	  }
	  break;
  }
  return 0;
}

void aff_part_buffer(const int newline,t_diskext_cst partition)
{
  switch(newline)
  {
    case 1: aff_buffer(BUFFER_ADD,"\n   "); break;
    case 2: aff_buffer(BUFFER_ADD,"%2d ", partition->order); break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           aff_buffer(BUFFER_ADD,"P"); break;
    case STATUS_PRIM_BOOT:      aff_buffer(BUFFER_ADD,"*"); break;
    case STATUS_EXT:            aff_buffer(BUFFER_ADD,"E"); break;
    case STATUS_EXT_IN_EXT:     aff_buffer(BUFFER_ADD,"X"); break;
    case STATUS_LOG:            aff_buffer(BUFFER_ADD,"L"); break;
    case STATUS_DELETED:        aff_buffer(BUFFER_ADD,"D"); break;
  }
  if(nom_os[partition->part_type])
    aff_buffer(BUFFER_ADD," %-28s ",nom_os[partition->part_type]);
  else
    aff_buffer(BUFFER_ADD," Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_buffer(&partition->start);
  aff_CHS_buffer(&partition->end);
  if(partition->sect_rel!=0)
    aff_buffer(BUFFER_ADD,"%10lu ", partition->part_size);
  aff_buffer(BUFFER_ADD,"%10lu", partition->part_size);
}

void aff_entree(t_entree_cst entree)
{
  if(nom_os[os(entree)])
    doprintf("\n %-28s ",nom_os[os(entree)]);
  else
    doprintf("\n Sys=%02X\t\t\t  ", os(entree));
  doprintf("%4u %3u %2u"
	 " %4u %3u %2u"
	 " %10lu"
	 " %10lu",
	  s_cyl(entree),s_head(entree),s_sect(entree),
	  e_cyl(entree),e_head(entree),e_sect(entree),
	  sect_rel(entree),nbr_sect(entree));
}

void aff_CHS_rapport(const t_CHS * CHS)
{
  ecrit_rapport("%4u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS(const t_CHS * CHS)
{
  doprintf("%4u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS_buffer(const t_CHS * CHS)
{
  aff_buffer(BUFFER_ADD,"%4u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_part_rapport(const int newline,t_diskext_cst partition)
{
  switch(newline)
  {
    case 1: ecrit_rapport("\n   "); break;
    case 2: ecrit_rapport("%2d ", partition->order); break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           ecrit_rapport("P"); break;
    case STATUS_PRIM_BOOT:      ecrit_rapport("*"); break;
    case STATUS_EXT:            ecrit_rapport("E"); break;
    case STATUS_EXT_IN_EXT:     ecrit_rapport("X"); break;
    case STATUS_LOG:            ecrit_rapport("L"); break;
    case STATUS_DELETED:        ecrit_rapport("D"); break;
  }
  if(nom_os[partition->part_type])
    ecrit_rapport(" %-28s ",nom_os[partition->part_type]);
  else
    ecrit_rapport(" Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_rapport(&partition->start);
  aff_CHS_rapport(&partition->end);
  if(partition->sect_rel!=0)
    ecrit_rapport("%10lu ", partition->part_size);
  ecrit_rapport("%10lu", partition->part_size);
}

void aff_part(const int newline,t_diskext_cst partition)
{
  switch(newline)
  {
    case 1: doprintf("\n   "); break;
    case 2: doprintf("%2d ", partition->order); break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           doprintf("P"); break;
    case STATUS_PRIM_BOOT:      doprintf("*"); break;
    case STATUS_EXT:            doprintf("E"); break;
    case STATUS_EXT_IN_EXT:     doprintf("X"); break;
    case STATUS_LOG:            doprintf("L"); break;
    case STATUS_DELETED:        doprintf("D"); break;
  }
  if(nom_os[partition->part_type])
    doprintf(" %-28s ",nom_os[partition->part_type]);
  else
    doprintf(" Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS(&partition->start);
  aff_CHS(&partition->end);
  if(partition->sect_rel!=0)
    doprintf("%10lu ", partition->part_size);
  doprintf("%10lu", partition->part_size);
  refresh();
}

void aff_LBA2CHS_rapport(t_param_disk_cst disk_car,dword pos_LBA)
{
  dword tmp;
  dword cylinder, head, sector;
  tmp=disk_car->CHS.sector;
  sector=(pos_LBA%tmp)+1;
  tmp=pos_LBA/tmp;
  cylinder=tmp/(disk_car->CHS.head+1);
  head=tmp%(disk_car->CHS.head+1);
  ecrit_rapport("%d/%d/%d", cylinder,head,sector);
}

void aff_disk(t_param_disk_cst param_disk)
{
#ifdef DJGPP
  doprintf(m_DRIVE " %02x - CHS %d %d %d - %d MB\n",
      param_disk->disk, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#else
  doprintf(m_DRIVE " %s - CHS %d %d %d - %d MB\n",
      param_disk->name, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#endif
}
